import sys, os

try:
    import setuptools
except ModuleNotFoundError:
    print("\n\tCan't import `setuptools`. Install `python3-setuptools` via APT or `setuptools` via `pip install`, etc. and re-run setup.")
    sys.exit(1)

# project information
name = 'vmelib'
version = '1.4'
author = "Viacheslav V. Kaminskiy, Stepan A. Zakharov"
author_email = "V.V.Kaminskiy@inp.nsk.su"
description = "C++ library ported into Python for communication with CAEN blocks via VME bus"
url = "https://git.inp.nsk.su/lsrp/vmelib"


# congig of extension
include_dirs = ["."]
library_dirs = []
extra_link_args = ["-Wl,-zdefs"]
libraries = ["CAENVME"]


def configure():
    print("`vmelib` setup configures libraries and paths.")
    print("Checking `CAENVME`...", end=" ")

    from ctypes.util import find_library
    from ctypes import __version__ as ctypes_ver
    ctypes_sees_env = ctypes_ver >= "3.6.0"
    caenvme_headers_dir = ""
    caenvme_lib_dir = ""
    lib = find_library("CAENVME")
    ld_library_path = os.getenv("LD_LIBRARY_PATH", default="")

    if lib or (ld_library_path and not ctypes_sees_env):
        print("`libCAENVME.so` was found in paths:", lib)
        #   try to search headers in default path and near `LD_LIBRARY_PATH`
        searchpaths = ["/usr/lib/x64/"] if os.path.exists(os.path.join("/usr/lib/", "libCAENVME.so")) else []
        if ld_library_path:
            searchpaths.extend(ld_library_path.split(":"))
        #   print(searchpaths)
        ans = "1"
        for p in searchpaths:
            headers_path = os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(p))), "include")
        #   print(headers_path)
            if os.path.exists(os.path.join(headers_path, "CAENVMElib.h")) and \
               os.path.exists(os.path.join(headers_path, "CAENVMEtypes.h")) and \
               os.path.exists(os.path.join(headers_path, "CAENVMEoslib.h")):
                ans = headers_path
                break
        while not os.path.exists(os.path.join(ans, "CAENVMElib.h")) or \
              not os.path.exists(os.path.join(ans, "CAENVMEtypes.h")) or \
              not os.path.exists(os.path.join(ans, "CAENVMEoslib.h")):
            ans = input("\tEnter path where `CAENVMElib.h`, `CAENVMEtypes.h` and `CAENVMEoslib.h` are located:\n\t\t")
        include_dirs.append(ans)
    else:
        print("`libCAENVME.so` was not found")
        ans = "1"
        while ans and not set("yYnN").intersection(ans):
            ans = input("\tDo you wish to hard-link libCAENVME to vmelib? If not, you have to use `LD_LIBRARY_PATH` in compile and run time. [y/N] ")
        if not ans or set("nN").intersection(ans):
            print("\tTo proceed, add path where `libCAENVME.so` is located to system variable `LD_LIBRARY_PATH` and re-run setup.")
            # where are headers??? look in <PATH>/../../include
            sys.exit(1)
        else:
            ans = " "
            while not os.path.exists(os.path.join(ans, "libCAENVME.so")):
                ans = input("\tEnter path where `libCAENVME.so` is located:\n\t\t")
            library_dirs.append(ans)
            ans = os.path.join(os.path.dirname(os.path.dirname(ans)), "include")
            while not os.path.exists(os.path.join(ans, "CAENVMElib.h")) or \
                  not os.path.exists(os.path.join(ans, "CAENVMEtypes.h")) or \
                  not os.path.exists(os.path.join(ans, "CAENVMEoslib.h")):
                ans = input("\tEnter path where `CAENVMElib.h`, `CAENVMEtypes.h` and `CAENVMEoslib.h` are located:\n\t\t")
            include_dirs.append(ans)


    # find Python version
    print("Checking Python development files...", end=" ")
    lib = "python{:d}.{:d}".format(sys.version_info.major, sys.version_info.minor)
    if   find_library(lib):
        libraries.append(lib)
        print("success")
    elif find_library("python3"): # libpython3 is stub in Linux Mint?
        libraries.append("python3")
    else:
        print("`libpython3*` and `Python.h` were not found.\n\tInstall `python3-dev` package and re-run setup.")
        sys.exit(1)

    # check numpy NOW PROCESSED BY SETUP
    '''
    print("Checking `numpy`...", end=" ")
    try:
        import numpy
        print("success")
    except ModuleNotFoundError:
        print("`numpy.h` was not found.\nInstall `python3-numpy-dev` via APT or `numpy` via `pip install`, etc. and re-run setup.")
        sys.exit(1)
    '''

if __name__ == "__main__":
    configure()
    c_caen_vme_device = setuptools.Extension("CCAENVMEDevice",
                           sources=["vmelib.cpp", "vmelib_wrap.cpp"],
                           extra_compile_args=["-DLINUX", "-std=c++11", "-s"],
                           include_dirs=include_dirs,
                           libraries=libraries,
                           library_dirs=library_dirs,
                           optional=False,
                           extra_link_args=extra_link_args
    )


    with open("README.rst", "r") as fh:
        long_description = fh.read()


    setuptools.setup(
        name=name,
        version=version,
        author=author,
        author_email=author_email,
        description=description,
        long_description=long_description,
        url=url,
    #    project_urls={
    #       "Bug Tracker": "https://example.com",
    #    },
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: POSIX :: Linux",
        ],
        py_modules=["vmelib", "vmelib_const"],
        package_dir={"" : "."},
        packages=setuptools.find_packages(where="."),
        python_requires=">=3.5",
        install_requires=["numpy"],
        ext_modules=[c_caen_vme_device],
    )

