'''
Name:         vmelib
Version:      1.3
Content:      Tools to operate with scientific devices using VME bus, e.g. CAEN VME crates
Authors:      V.V. Kaminskiy, S.A. Zakharov
Organization: Budker Institute of Nuclear Physics, Novosibirsk, Russian Federation
E-mail:       V.V.Kaminskiy@inp.nsk.su

'''
try:
    import CCAENVMEDevice
except:
    raise RuntimeError('''Can not import `CCAENVMEDevice` module!
    1. Check `libCAENVME.so` location: `/usr/lib` (if in was installed as written in `CAENVMElib` docs),
       or location defined in `LD_LIBRARY_PATH` variable (at build and run time).
    2. Check if `CCAENVMEDevice.cpython-3*-*-*.so` exists in `vmelib` installation directory.
       If absent, re-install `vmelib`.''')
    
from vmelib_const import *


class CAENVMEDevice():
   """
An object representing CAEN VME crate with communication bridge. It allows to perform read/write/irq operations of VME devices (modules).

"""

   def __init__(self, bd_type=V2718, link=0, bd_num=0, addr_mode=A32_S_DATA, data_width=D32, verbosity=0, base_addr=0):
      """
Opens communication board and sets base address and default communication parameters.

If you use one VME module, it is convenient to set all the parameters in `__init__`, including base address. Otherwise, if using few links/boards, you should init the device without module-specific parameters and the set link/board with `open_board`.

Parameters
----------
bd_type:    int, optional
    Model of the bridge for further data transfers. Can be modified further in `open_board`. Default: `vmelib.V2718`.
link:       int, optional
    Link number (don't care for V1718) for further data transfers, e.g, number of PCI card. Can be modified further in `open_board`. Default: 0.
bd_num:     int, optional
    Board number in the link for further data transfers, e.g, number of slot in PCI card. Can be modified further in `open_board`. Default: 0.
addr_mode:  int, optional
    Address modifier. This value is set as default in read/write functions and can be modified in `open_device`. Default: `vmelib.A32_S_DATA`.
data_width: int, optional
    Data chunk size, in bytes. This value is set as default in read/write functions and can be modified in `open_device`. Default: `vmelib.D32`.
verbosity:  int, optional
    0 - debug messages are not printed. 1 - some debug messages are printed in stdout. 2 - all debug messages are printed in stdout. Default: 0.
base_addr:  int, optional
    VME base address in format 0xXXXX0000. This value is set as default in read/write functions and can be modified in `open_device`. Default: 0xFFFFFFFF.

Returns
-------
:obj:`CAENVMEDevice`
    An instance of "VME crate".

Raises
------
VMElibError 
    If error occurs.

"""
      self.CCAENVMEDevice_capsule = CCAENVMEDevice.construct(bd_type, link, bd_num, addr_mode, data_width, verbosity, base_addr)


   def __del__(self):
      """`CAENVMEDevice` destructor."""
      CCAENVMEDevice.delete_object(self.CCAENVMEDevice_capsule)


   def open_board(self, bd_type, link, bd_num):
      """
Opens communication board. It can useful if you want open another board in the same `CAENVMEDevice` instance.
      
Parameters
----------
bd_type: int
    model of the bridge, e.g. `vmelib.V2718`.
link:   int
    link number (don't care for V1718), e.g, number of PCI card, default: 0.
bd_num: int
    board number in the link, e.g, number of slot in PCI card, default: 0.

Returns
-------
None
Raises
------
VMElibError 
    If error occurs.

"""
      self.__error__(CCAENVMEDevice.open_board(self.CCAENVMEDevice_capsule, bd_type, link, bd_num))

     
   def open_device(self, base_addr, addr_mode, data_width):
      """
Sets default communication parameters for VME device (module). Does not make any data transfer. It is convenient when you work with one module.

Parameters
----------
base_addr:  int
    VME base address in format 0xXXXX0000, arbitrary. This value is set as default in proceeding read/write functions.
addr_mode:  int
    Address modifier (e.g, vmelib.A32_S_DATA). This value is set as default in read/write functions.
data_width: int
    Data chunk size, in bytes (e.g, vmelib.D32), arbitrary. This value is set as default in read/write functions.

Returns
-------
None
"""
      self.__error__(CCAENVMEDevice.open_device(self.CCAENVMEDevice_capsule, base_addr, addr_mode, data_width))


############# SINGLE REGISTERS ###############
   def read_reg(self, reg_addr, base_addr=INVALID_BA, addr_mode=INVALID_MODE, data_width=INVALID_MODE):
      """
Reads single register from VME device (module).

Parameters
----------
reg_addr:   int
    Address of the register in the device in format 0xXXXX.
base_addr:  int, optional
    VME base address in format 0xXXXX0000. If not set, the default value is used (which was set earlier by `__init__` or `open_device`).
addr_mode:  int, optional
    Adress modifier. If not set, the default value is used (which was set earlier by `__init__` or `open_device`)
data_width: int, optional
    Data chunk size, in bytes. If not set, the default value is used (which was set earlier by `__init__` or `open_device`).

Returns
-------
int
    Value read.

Raises
------
VMElibError 
    If error occurs.

"""
      err_code, data = CCAENVMEDevice.read_reg(self.CCAENVMEDevice_capsule, reg_addr, base_addr, addr_mode, data_width)
      self.__error__(err_code)
      return data


   def write_reg(self, data, reg_addr, base_addr=INVALID_BA, addr_mode=INVALID_MODE, data_width=INVALID_MODE):
      """
Writes single register to VME device (module).

Parameters
----------
data:       int
    Data to write.
reg_addr:   int
    Address of the register in thew device in format 0xXXXX.
base_addr:  int, optional
    VME base address in format 0xXXXX0000. If not set, the default value is used (which was set earlier by `__init__` or `open_device`).
addr_mode:  int, optional
    Address modifier. If not set, the default value is used (which was set earlier by `__init__` or `open_device`).
data_width: int, optional
    Data chunk size, in bytes. If not set, the default value is used (which was set earlier by `__init__` or `open_device`).

Returns
-------
None

Raises
------
VMElibError 
    If error occurs.

"""
      self.__error__(CCAENVMEDevice.write_reg(self.CCAENVMEDevice_capsule, data, reg_addr, base_addr, addr_mode, data_width))


   def read_write(self, data, reg_addr, base_addr=INVALID_BA, addr_mode=INVALID_MODE, data_width=INVALID_MODE):
      """
Reads single register to VME device (module) and immediately the writes data into it.

Parameters
----------
data:       int
    Data to write.
reg_addr:   int
    Address of the register in thew device in format 0xXXXX.
base_addr:  int, optional
    VME base address in format 0xXXXX0000. If not set, the default value is used (which was set earlier by `__init__` or `open_device`).
addr_mode:  int, optional
    Address modifier. If not set, the default value is used (which was set earlier by `__init__` or `open_device`).
data_width: int, optional
    Data chunk size, in bytes. If not set, the default value is used (which was set earlier by `__init__` or `open_device`).

Returns
-------
int
    Value read.

Raises
------
VMElibError 
    If error occurs.

"""
      err_code, reg = CCAENVMEDevice.read_write(self.CCAENVMEDevice_capsule, data, reg_addr, base_addr, addr_mode, data_width)
      self.__error__(err_code)
      return reg


############# MANY REGISTERS ###############
   def read_page(self, size, reg_addr, base_addr=INVALID_BA, addr_mode=INVALID_MODE, data_width=INVALID_MODE, blt=True):
      """
Reads many contiguous registers (page) from VME device (module). It can perform both block transfer (BLT) read and single read of contiguous registers in single data transfer cycle.

Parameters
----------
size:       int
    Size of array to read, i.e, number of values (not bytes!).
reg_addr:   int
    Address of the first register to read in the device in format 0xXXXX.
base_addr:  int, optional
    VME base address in format 0xXXXX0000. If not set, the default value is used (which was set earlier by `__init__` or `open_device`).
addr_mode:  int, optional
    Address modifier. Address modifier can be set for single register access (i.e, `vmelib.A32_U_DATA` instead of `vmelib.A32_U_BLT`), it is automatically corrected. If not set, the default value is used (which was set earlier by `__init__` or `open_device`).
data_width: int, optional
    Data chunk size, in bytes. If not set, the default value is used (which was set earlier by `__init__` or `open_device`).
blt:        bool, optional
    If `True` performs real BLT read cycle (faster), if `False` performs many single reads in single data transfer cycle (slower), default: `True` (do BLT).
 
Returns
-------
:obj: `numpy.array`
    Values read.

Raises
------
VMElibError 
    If error occurs.

See Also
--------
read_multi
    to read non-contiguous registers, it is slower. `read_page` do the same type of data access if `blt` is `False`.

"""
      err_code, data = CCAENVMEDevice.read_page(self.CCAENVMEDevice_capsule, size, reg_addr, base_addr, addr_mode, data_width, blt)
      self.__error__(err_code)
      return data


   def write_page(self, data, reg_addr, base_addr=INVALID_BA, addr_mode=INVALID_MODE, data_width=INVALID_MODE, blt=True):
      """
Write many contiguous registers (page) from VME device (module). It can perform both block transfer (BLT) write and single write of contiguous registers in single data transfer cycle.

Parameters
----------
data:       list of int
    Data to write.
reg_addr:   int
    Address of the first register to write in the device in format 0xXXXX.
base_addr:  int, optional
    VME base address in format 0xXXXX0000. If not set, the default value is used (which was set earlier by `__init__` or `open_device`).
addr_mode:  int, optional
    Address modifier. Address modifier can be set for single register access (i.e, `vmelib.A32_U_DATA` instead of `vmelib.A32_U_BLT`), it is automatically corrected. If not set, the default value is used (which was set earlier by `__init__` or `open_device`).
data_width: int, optional
    Data chunk size, in bytes. If not set, the default value is used (which was set earlier by `__init__` or `open_device`).
blt:        bool, optional
    If `True` performs real BLT write cycle (faster), if `False` performs many single writes in single data transfer cycle (slower), default: `True` (BLT).
 
Returns
-------
None

Raises
------
VMElibError 
    If error occurs.

See Also
--------
write_multi
    to write non-contiguous registers, it is slower. `write_page` do the same type of data access if `blt` is `False`.

"""
      self.__error__(CCAENVMEDevice.write_page(self.CCAENVMEDevice_capsule, data, reg_addr, base_addr, addr_mode, data_width, blt))


   def read_multi(self, regs, addr_modes=[], data_widths=[]):
      """
Reads many arbitrary registers from VME device (module), i.e, perform many single reads.

Parameters
----------
regs:       list of int
    Addresses of the registers in format 0xYYYYXXXX (register address + base address) or 0xXXXX (register address only, default base address is used).
addr_modes: list of int, optional
    Address modifiers. If the list is empty, the default values are used (which was set earlier in `__init__` or `open_device`). Default: empty list.
data_widths: list of int, optional
    Data chunk sizes, in bytes. If the list is empty, the default values are used (which was set earlier in `__init__` or `open_device`). Default: empty list.

Returns
-------
:obj: `numpy.array`
    Values read.

Raises
------
VMElibError 
    If error occurs.
IndexError
    If `addr_mode` or `data_width` have wrong length.

See Also
--------
read_page
    to read contiguous registers, it is faster in BLT mode.

"""
      size = len(regs)
      if not addr_modes:
         addr_modes = [INVALID_MODE] * size
      if not data_widths:
         data_widths = [INVALID_MODE] * size
      if "__len__" not in dir(addr_modes) or "__len__" not in dir(data_widths) or \
                    len(addr_modes)!=size or len(data_widths)!=size:
         raise IndexError("`addr_modes` and `data_widths` must be list-like objects of the `regs` length or None (empty list, False) objects.")
      err_code, data = CCAENVMEDevice.read_multi(self.CCAENVMEDevice_capsule, regs, addr_modes, data_widths)
      self.__error__(err_code)
      return data


   def write_multi(self, data, regs, addr_modes=[], data_widths=[]):
      """
Write many arbitrary registers from VME device (module), i.e, perform many single writes.

Parameters
----------
data:       list of int
    Data to write.
regs:       list of int
    Addresses of the registers in format 0xYYYYXXXX (register address + base address) or 0xXXXX (register address only, default base address is used).
addr_modes: list of int, optional
    Address modifiers. If the list is empty, the default values are used (which was set earlier in `__init__` or `open_device`). Default: empty list.
data_widths: list of int, optional
    Data chunk sizes, in bytes. If the list is empty, the default values are used (which was set earlier in `__init__` or `open_device`). Default: empty list.

Returns
-------
None

Raises
------
VMElibError 
    If error occurs.
IndexError
    If `data`, `regs`, `addr_mode` or `data_width` have wrong lengths.

See Also
--------
write_page
    to write contiguous registers, it is faster in BLT mode.

"""
      size = len(regs)
      
      if len(data)!=size:
         raise IndexError("`data` and `regs` must be the same length.")
      if not addr_modes:
         addr_modes = [INVALID_MODE] * size
      if not data_widths:
         data_widths = [INVALID_MODE] * size
      if "__len__" not in dir(addr_modes) or "__len__" not in dir(data_widths) or \
                    len(addr_modes)!=size or len(data_widths)!=size:
         raise IndexError("`addr_modes` and `data_widths` must be list-like objects of the `regs` length or None (empy list, False) objects.")
      err_code = CCAENVMEDevice.write_multi(self.CCAENVMEDevice_capsule, data, regs, addr_modes, data_widths)
      self.__error__(err_code)


############# IRQ STUFF ###############
   def irq_enable(self, mask):
      """
Enables the IRQ lines specified by mask.

Parameters
----------
mask:       int
    Bit mask: interrupt levels from 1 to 7 (from IRQ1 to IRQ7).

Returns
-------
None

Raises
------
VMElibError 
    If error occurs.

"""
      self.__error__(CCAENVMEDevice.irq_enable(self.CCAENVMEDevice_capsule, mask))


   def irq_disable(self, mask):
      """Disables the IRQ lines specified by mask.

Parameters
----------
mask:       int
    Bit mask: interrupt levels from 1 to 7 (from IRQ1 to IRQ7).

Returns
-------
None

Raises
------
VMElibError 
    If error occurs.

"""
      self.__error__(CCAENVMEDevice.irq_disable(self.CCAENVMEDevice_capsule, mask))


   def irq_wait(self, mask, timeout):
      """Waits the IRQ lines specified by mask during timeout.

Parameters
----------
mask:       int
    Bit mask: interrupt levels from 1 to 7 (from IRQ1 to IRQ7).
timeout:    int
    Timeout in milliseconds.

Returns
-------
None

Raises
------
VMElibError("TimeoutError")
    If timeout expires.
VMElibError 
    If other error occurs.

"""
      self.__error__(CCAENVMEDevice.irq_wait(self.CCAENVMEDevice_capsule, mask, timeout))


   def irq_check(self):
      """Check the IRQ lines.

Parameters
----------
None

Returns
-------
list of int
    Bit mask: interrupt levels from 1 to 7 (from IRQ1 to IRQ7).

Raises
------
VMElibError 
    If error occurs.

See Also
--------
irq_wait:
    faster and more convenient way to follow IRQ state.

"""
      err_code, mask = CCAENVMEDevice.irq_check(self.CCAENVMEDevice_capsule)
      self.__error__(err_code)
      return [(mask>>b) & 1 for b in range(7, -1, -1)]


#############################################
   def get_timeout(self):
      """Get VME bus timeout.

Parameters
----------
None

Returns
-------
int
    VME bus timeout: 50 us or 400 us.

Raises
------
VMElibError 
    If error occurs.

"""
      err_code, tm = CCAENVMEDevice.get_timeout(self.CCAENVMEDevice_capsule)
      self.__error__(err_code)
      return tm


#############################################
   def set_timeout(self, timeout):
      """Set VME bus timeout.

Parameters
----------
timeout:    int
    Timeout in microseconds, 50 or 400.

Returns
-------
None

Raises
------
VMElibError 
    If error occurs.

"""
      self.__error__(CCAENVMEDevice.set_timeout(self.CCAENVMEDevice_capsule, timeout))


#############################################
   def get_arbiter_type(self):
      """Get the type of VME bus arbiter implemented on the module.

Parameters
----------
None

Returns
-------
str
    Arbiter type: "priority arbiter" or "round-robin arbiter".

Raises
------
VMElibError 
    If error occurs.

"""
      err_code, arb = CCAENVMEDevice.get_arbiter_type(self.CCAENVMEDevice_capsule)
      self.__error__(err_code)
      if   arb == ArbiterPriorized:
         return "priority arbiter"
      elif arb == ArbiterRoundRobin:
         return "round-robin arbiter"


############################################
   def set_arbiter_type(self, arbiter_type):
      """Set the type of VME bus arbiter implemented on the module.

Parameters
----------
arbiter_type: int or str
    Arbiter type: 0 ("priority arbiter") or 1 ("round-robin arbiter").

Returns
-------
None

Raises
------
ValueError
    If `arbiter_type` is not integer or string in ["priority arbiter", "round-robin arbiter"].

VMElibError 
    If error occurs.

"""
      arb = None
      if   arbiter_type == "priority arbiter":
         arb = ArbiterPriorized
      elif arbiter_type == "round-robin arbiter":
         arb = ArbiterRoundRobin
      elif arbiter_type in (ArbiterPriorized, ArbiterRoundRobin):
         arb = arbiter_type
      if arb is None:
         raise ValueError('`arbiter_type` must be 0 (or "priority arbiter") or 1 (or "round-robin arbiter")')

      self.__error__(CCAENVMEDevice.set_arbiter_type(self.CCAENVMEDevice_capsule, arb))


##########################
   def get_requester_type(self):
      """Get VME requester type.

Parameters
----------
None

Returns
-------
str
    VME bus requester type: "fair" (fair bus requester) or "demand" (on demand bus requester).

Raises
------
VMElibError 
    If error occurs.

"""
      err_code, requester_type = CCAENVMEDevice.get_requester_type(self.CCAENVMEDevice_capsule)
      self.__error__(err_code)
      if   requester_type == RequesterFair:
         return "fair"
      elif requester_type == RequesterDemand:
         return "demand"


#####################################
   def set_requester_type(self, requester_type):
      """Set VME requester type.

Parameters
----------
requester_type: int or str
    VME bus requester type: 0 ("fair", fair bus requester) or 1 ("demand", on demand bus requester).

Returns
-------
None

Raises
------

ValueError
    If `requester_type` not in [0, 1, "fair", "demand"].

VMElibError 
    If error occurs.

"""
      req = None
      if   requester_type == "fair":
         req = RequesterFair
      elif requester_type == "demand":
         req = RequesterDemand
      elif requester_type in (RequesterFair, RequesterDemand):
         req = requester_type
      if req is None:
         raise ValueError('`requester_type` must be 0 ("fair") or 1 ("demand")')
      self.__error__(CCAENVMEDevice.set_requester_type(self.CCAENVMEDevice_capsule, req))


###########################
   def get_requester_level(self):
      """Get VME requester priority level.

Parameters
----------
None

Returns
-------
int
    VME bus requester priority level: 0, 1, 2, 3.

Raises
------
VMElibError 
    If error occurs.

"""
      err_code, requester_level = CCAENVMEDevice.get_requester_level(self.CCAENVMEDevice_capsule)
      self.__error__(err_code)
      return requester_level


######################################
   def set_requester_level(self, requester_level):
      """Set VME requester priority level.

Parameters
----------
requester_level: int
    VME bus requester priority level: 0, 1, 2, 3.

Returns
-------
None

Raises
------

ValueError
    If `requester_level` not in [0, 1, 2, 3].

VMElibError 
    If error occurs.

"""
      if requester_level not in range(4):
         raise ValueError("`requester_level` must be in [0, 1, 2, 3]")
      self.__error__(CCAENVMEDevice.set_requester_level(self.CCAENVMEDevice_capsule, requester_level))


##########################
   def get_release_type(self):
      """Get VME bus release type.

Parameters
----------
None

Returns
-------
str
    VME bus release type: "RWD" (Release When Done) or "ROR" (Release On Request).

Raises
------
VMElibError 
    If error occurs.

"""
      err_code, release_type = CCAENVMEDevice.get_release_type(self.CCAENVMEDevice_capsule)
      self.__error__(err_code)
      if   release_type == ReleaseRWD:
         return "RWD"
      elif release_type == ReleaseROR:
         return "ROR"


#####################################
   def set_release_type(self, release_type):
      """Set VME bus release type.

Parameters
----------
release_type: int or str
    VME bus release type: 0 (or "RWD", Release When Done) or 1 (or "ROR", Release On Request).

Returns
-------
None

Raises
------

ValueError
    If `release_type` not in [0, 1, "RWD", "ROR"].

VMElibError 
    If error occurs.

"""
      rel = None
      if   release_type == "RWD":
         rel = ReleaseRWD
      elif release_type == "ROR":
         rel = ReleaseROR
      elif release_type in (ReleaseRWD, ReleaseROR):
         rel = release_type
      if rel is None:
         raise ValueError('`release_type` must be 0 ("RWD") or 1 ("ROR")')
      self.__error__(CCAENVMEDevice.set_requester_type(self.CCAENVMEDevice_capsule, rel))


#############################################
   def get_sw_info(self):
      """Get software and firmware versions.

Parameters
----------
None

Returns
-------
dict of str:
    `CAENVMElib` version, board (bridge) firmware version.

Raises
------
VMElibError 
    If error occurs.

"""
      err_code, ver = CCAENVMEDevice.get_sw_info(self.CCAENVMEDevice_capsule)
      self.__error__(err_code)
      return {k : v.strip() for k,v in zip(["Library version", "Board FW version"], ver.split("|"))}




############# GENERATE ERRORS ###############
   def __error__(self, err_code, bd_type=None, link=None, bd_num=None, base_addr=None, reg_addr=None):
     """
Converts error codes returned by C functions to Python exceptions.

Parameters
----------
err_code:   int
    Error code returned by C CAENVMElib functions.
bd_type, link, bd_num, base_addr, reg_addr: int
    See :meth:`__init__` for detailed description.

Returns
-------
None

Raises
------
VMElibError
    If `err_code` is non-zero.

"""
     if err_code == Success:
        pass # Operation completed successfully
     elif err_code == BusError:
        raise VMElibError("BusError", "VME bus error during the cycle")
     elif err_code == CommError:
        raise VMElibError("CommError", "Communication error") 
     elif err_code == GenericError:
        raise VMElibError("GenericError", "Unspecified error")
     elif err_code == InvalidParam:
        raise VMElibError("InvalidParam", "Invalid parameter")
     elif err_code == TimeoutError:
        raise VMElibError("TimeoutError", "Timeout error")
     elif err_code == AlreadyOpenError:
        raise VMElibError("AlreadyOpenError", "Device is already open error")
     elif err_code == MaxBoardCountError:
        raise VMElibError("MaxBoardCountError", "Maximum device number has been reached")
     elif err_code == NotSupported:
        raise VMElibError("NotSupported", "Function not supported by that board model")
     else:
        raise VMElibError("UnknownError", "Unknown error")



class VMElibError (Exception):
   '''
Exceptions raised on `CAENVMElib` errors.
'''
   def __init__(self, err_type, name, bd_type=None, link=None, bd_num=None,
                base_addr=None, reg_addr=None, action=None):
      self.err_type = err_type
      self.name = name
      self.bd_type = bd_type
      self.link = link
      self.bd_num = bd_num
      self.base_addr = base_addr
      self.reg_addr = reg_addr
      self.action = action
   

   def make_message(self):
      '''
Make human-readable message on error.
      '''
      mesg = []
      if bd_type is not None:
         mesg.append("board: {}".format(bd_type))
      if link is not None:
         mesg.append("link: {}".format(link))
      if bd_num is not None:
         mesg.append("board #: {}".format(bd_num))
      if base_addr is not None:
         mesg.append("base address: {}".format(base_addr))
      if reg_addr is not None:
         mesg.append("register address: {}".format(reg_addr))
      
      print(self.message, " while ", end="")
      if mesg:
         if self.action is not None:
            print(action, end="")
         else:
            print("accessing")
         print(", ".join(mesg), ".")



###### EXAMPLE (QDC) #####
if __name__=="__main__":
   dev = CAENVMEDevice(verbosity=2)
   dev.open_device(0x32110000, A32_S_DATA, D16)
   # clear bit # 14 (ALL_TRG)
   dev.write_reg(0x4000, 0x1034)
   #  set bits # 3 (OVER_RANGE_EN) and # 2 (CLEAR_DATA)
   dev.write_reg(0x10 | 0x8 | 0x4, 0x1032)
   # clear bit # 2 (CLEAR_DATA)
   dev.write_reg(0x4, 0x1034)
   # kill channels and set thresholds
   for reg in range(0x1080, 0x10BF, 4):
      if reg == 0x10b8:
         dev.write_reg(180>>4, reg)
      else:
         # kill channel
         dev.write_reg(0x100, reg)
   # enable interrupt generation: IRQ level 1
   dev.write_reg(0x1, 0x100A)
   # set buffer interrupt generation at N events
   dev.write_reg(20, 0x1020)
   n_total = 0
   # clear buffer
   #   set bit # 2 (CLEAR_DATA)
   dev.write_reg(0x4, 0x1032)
   # clear bit # 2 (CLEAR_DATA)
   dev.write_reg(0x4, 0x1034)
   dev.irq_enable(0x1)
   words = dev.read_page(100, 0, data_width=D32)
   print(words)

